(ns document-storage.postgres.core
  (:require [clojure.tools.logging :as log]
            [clojure.java.io :as io]
            [clojure.edn :refer [read-string]]
            [clojure.set :refer [difference]]
            [fipp.edn :refer [pprint] :rename {pprint fipp}]
            [document-storage.protocol :as protocol]
            [document-storage.postgres.db :as db]
            [document-storage.postgres.documents :as documents]
            [document-storage.query :refer [query-fits? update-queries
                                            update-queries-remove-document]]
            [conman.core :as conman]
            [clj-helper.string :refer [get-unique-id]]))

(defprotocol db-storage
  "Base Protocol for document storages"
  (get-schema [this] "Return the db-schema name")
  (create-repository [this repository] "Create a new repository")
  (delete-repository [this repository] "Delete a repository")
  (run-query [this query-key params] "Run a Query on a DB-based Storage"))

(defrecord postgres-storage [storage-id params connection queries]
  db-storage
  (get-schema [this]
    (.-storage-id this))
  (create-repository [this repository]
    (.run-query this :create-repository! {:repository repository}))
  (delete-repository [this repository]
    (.run-query this :delete-repository! {:repository repository}))
  (run-query [this query-key params]
    (let [storage-schema (.get-schema this)
          repository (:repository params)
          repository-name (if (keyword? repository)
                            (name repository)
                            (str repository))
          table-names {:schema-name storage-schema
                       :repository-name repository-name
                       :ds-encoding-fqn (str storage-schema
                                           "."
                                           "ds_encoding")
                       :ds-type-fqn (str storage-schema
                                       "."
                                       "ds_type")
                       :repositories-table (str storage-schema
                                                "."
                                                "repositories")
                       :repository-table (str storage-schema
                                              "."
                                              "repo_"
                                              repository-name)
                       :uuid (get-unique-id "ds")}]
      (conman/query (:queries this) query-key
                    (merge table-names params))))

  protocol/storage
  (init-storage [this]
    (log/info "Init Postgres-Storage.")
    (.run-query this :create-storage! {})
    (let [repositories (set (map name (get (.-params this) :repositories [])))
          existing-repos (set (map :id (.list-repositories this)))]
      (log/info "Repositories: " repositories)
      (log/info "Existing Repositories: " existing-repos)
      (doseq [r (difference repositories existing-repos)]
        (log/info "Create Repository: " r)
        (.create-repository this r))))

  (list-repositories [this]
    (run-query this :get-repositories {}))

  (load-document [this id repository]
    (documents/load-document this id repository)
    )

  (get-document-size [this id repository]
    (count (documents/load-document this id repository)))

  (save-document [this id document repository]
    (documents/save-document this id document repository))

  (delete-document [this id repository]
    (documents/delete-document this id repository))

  (list-documents [this repository]
    (into #{}
          (map :document_id (run-query this :get-all-document-ids
                                       {:repository repository}))))

  (find-documents [this query repository]
    (let [docs (.list-documents this :repository repository)]
      (let [result (into {} (keep (fn [fname]
                                    (let [doc (.load-document this fname :repository repository)]
                                      (when (query-fits? doc query)
                                        [fname doc]
                                        )
                                      ))
                                  docs))]
        result)))

  (shutdown [this]
    (db/disconnect! (:connection this)))

  (list-versions [this id repository]
    (run-query this :get-document-versions {:repository repository
                                            :document-id id}))

  (load-version [this id version repository]
    (documents/load-document this id version repository))
  )

(defn make-postgres-ds [config]
  (let [conn (db/connect! config)
        id (str "ds" (name (or (:id config) "main")))
        storage (->postgres-storage id {:repositories (:repositories config)} (:connection conn) (:queries conn))]
    (.init-storage storage)
    storage))
